"""MDU Parser."""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Union

from hydrolib.core.basemodel import PathOrStr
from hydrolib.core.dflowfm.mdu.models import FMModel, Physics, Time
from hydrolib.tools.extforce_convert.utils import IgnoreUnknownKeyWordClass, backup_file


class MDUParser:
    """A class to update the ExtForceFileNew entry in an MDU file."""

    def __init__(self, mdu_path: PathOrStr):
        """Initialize the MDUParser.

        Args:
            mdu_path: Path to the MDU file to update
        """
        mdu_path = Path(mdu_path) if not isinstance(mdu_path, Path) else mdu_path
        if not mdu_path.exists():
            raise FileNotFoundError(f"File not found: {mdu_path}")

        self.mdu_path = mdu_path
        self.updated_lines = []
        self.inside_external_forcing = False
        self.found_extforcefilenew = False
        self._content = self._read_file()
        self.loaded_fm_data = self._load_with_fm_model()
        self.temperature_salinity_data = self.get_temperature_salinity_data()

    def _load_with_fm_model(self) -> Dict[str, Any]:
        """Load the MDU file using the FMModel class.

        Returns:
            data (Dict[str, str]):
                all the data inside the mdu file, with each section in the file as a key and the data of that section is
                the value of that key.
        """
        data = FMModel._load(FMModel, self.mdu_path)
        return data

    @property
    def new_forcing_file(self) -> Union[Path, str]:
        """Get the filename for the ExtForceFileNew entry."""
        if not hasattr(self, "_new_forcing_file"):
            raise AttributeError("new_forcing_file not set")
        return self._new_forcing_file

    @new_forcing_file.setter
    def new_forcing_file(self, file_name: Union[Path, str]) -> None:
        """Set the new filename for the ExtForceFileNew entry.

        Args:
            file_name(PathOrStr):
                New filename for the ExtForceFileNew entry
        """
        if not isinstance(file_name, (Path, str)):
            raise ValueError("new_forcing_filename must be a str or Path")
        self._new_forcing_file = file_name

    def _read_file(self) -> List[str]:
        """Read the MDU file into a list of strings.

        Returns:
            List of strings, one for each line in the file
        """
        with open(self.mdu_path, "r", encoding="utf-8") as f:
            lines = f.readlines()
        return lines

    def save(self, backup: bool = False) -> None:
        """Save the updated MDU file."""
        if backup:
            backup_file(self.mdu_path)

        with open(self.mdu_path, "w", encoding="utf-8") as file:
            file.writelines(self.content)

    @property
    def content(self) -> List[str]:
        """Get the content of the MDU file.

        Returns:
            List of strings, one for each line in the file
        """
        return self._content

    @content.setter
    def content(self, new_content: List[str]) -> None:
        """Set the content of the MDU file.

        Args:
            new_content: New content for the MDU file
        """
        self._content = new_content

    def update_extforce_file_new(self) -> List[str]:
        """Update the ExtForceFileNew entry in the MDU file.

        Update the 'ExtForceFileNew' entry under the '[external forcing]' section
        of an MDU file. Writes the updated content to output_path if provided,
        or overwrites the original file otherwise.

        Returns:
            List[str]:
                The updated lines of the .mdu file.

        Notes:
            - This function is a workaround for updating the ExtForceFileNew entry in an MDU file.
            - The function reads the entire file into memory, updates the line containing ExtForceFileNew, and writes the
                updated content back to disk.
            - The function removes the `extforcefile` from the mdu file, and only keeps the new updated `ExtForceFileNew`
            entry, as all the old forcing quantities in the old forcing file are converted to the new format.
            - after fixing the issue with mdu files having `Unkown keyword` error, this function will be removed,
            and the `LegacyFMModel` will be the only way to read/update the mdu file

        """
        lines = self.content
        for line in lines:
            stripped = line.strip()
            # Check if we've hit the [external forcing] header
            if stripped.lower().startswith("[external forcing]"):
                self.inside_external_forcing = True
                self.found_extforcefilenew = False
                self.updated_lines.append(line)
                continue

            # If we are inside the [external forcing] section, look for ExtForceFileNew
            if self.inside_external_forcing:
                # If we find another section header, it means [external forcing] section ended
                self._handle_external_forcing_section(stripped)
                continue

            # Default: write the line unmodified
            self.updated_lines.append(line)

        # If we ended the file while still in [external forcing] with no ExtForceFileNew found, add it
        if self.inside_external_forcing and not self.found_extforcefilenew:
            new_line = (
                f"ExtForceFileNew                           = {self.new_forcing_file}\n"
            )
            self.updated_lines.append(new_line)
            self.updated_lines.append("\n")

        return self.updated_lines

    @staticmethod
    def is_section_header(line: str) -> bool:
        """Check if the line is a section header (e.g., '[...]').

        Args:
            line: The line to check

        Returns:
            True if the line is a section header (excluding '[external forcing]'), False otherwise
        """
        return (
            line.startswith("[")
            and line.endswith("]")
            and "external forcing" not in line.lower()
        )

    def replace_extforcefilenew(self, line: str) -> str:
        """Replace the ExtForceFileNew line with the new filename.

        Args:
            line: The line containing the ExtForceFileNew entry

        Returns:
            The updated line with the new filename

        Notes:
            - This method preserves the formatting of the original line, including comments
            - If the line doesn't contain an '=' character, it's returned unchanged
            - If the new filename would overflow into a comment, the comment is preserved
        """
        # Find the '=' character
        eq_index = line.find("=")
        if eq_index == -1:
            return line
        # Everything up to and including '='
        left_part = line[: eq_index + 1]
        # Remainder of the line (after '=')
        right_part = line[eq_index + 1 :].strip("\n")  # noqa: E203
        name_len = len(str(self.new_forcing_file))
        # Protect against filename overflow into the comment
        right_part_clipped = right_part[name_len + 1 :]
        if right_part_clipped.find("#") == -1 and right_part_clipped != "":
            right_part_clipped = f" {right_part.lstrip()}"
        # Insert new filename immediately after '=' + a space
        return f"{left_part} {self.new_forcing_file}{right_part_clipped}\n"

    def _handle_external_forcing_section(self, stripped: str) -> None:
        """Handle a line within the [external forcing] section.

        Args:
            stripped: The stripped line to process

        Notes:
            - If the line is a section header, it means we're leaving the [external forcing] section
            - If we're leaving the section and haven't found ExtForceFileNew, add it
            - If the line starts with ExtForceFileNew, replace it with the new filename
            - If the line starts with ExtForceFile, skip it (remove it from the output)
            - Otherwise, add the line to the output unchanged
        """
        if self.is_section_header(stripped):
            # If we never found ExtForceFileNew before leaving, add it now
            if not self.found_extforcefilenew:
                new_line = f"ExtForceFileNew                           = {self.new_forcing_file}\n"
                self.updated_lines.extend([new_line, "\n"])

            self.inside_external_forcing = False
            # fall through to just append the line below

        # If the line has ExtForceFileNew, replace it
        # The simplest way is to check if it starts with or contains ExtForceFileNew
        # ignoring trailing spaces. You can refine the logic as needed.
        if stripped.lower().startswith("extforcefilenew"):
            self.found_extforcefilenew = True
            self.updated_lines.append(self.replace_extforcefilenew(stripped))
            return
        elif stripped.lower().startswith("extforcefile"):
            return
        self.updated_lines.append(f"{stripped}\n")

    def get_temperature_salinity_data(self) -> Dict[str, Any]:
        """Get the info needed from the mdu to process and convert the old external forcing files.

        Returns:
            temperature_and_salinity_info (Dict[str, str]):
                dictionary with the information needed for the conversion tool to convert the `SourceSink` and
                `Boundary` quantities. The dictionary will have three keys `temperature`, `salinity`, and `refdate`.

        Examples:
            ```python
            >>> mdu_info = { #doctest: +SKIP
            ...     "refdate": "minutes since 2015-01-01 00:00:00",
            ...     "temperature": True,
            ...     "salinity": True,
            ... }
            ```
        """
        # read sections of the mdu file.
        time_data = self.loaded_fm_data.get("time")
        physics_data = self.loaded_fm_data.get("physics")
        mdu_time = IgnoreUnknownKeyWordClass(Time, **time_data)
        mdu_physics = IgnoreUnknownKeyWordClass(Physics, **physics_data)

        ref_time = get_ref_time(mdu_time.refdate)
        temperature_and_salinity_info = {
            "file_path": self.mdu_path,
            "refdate": ref_time,
            "temperature": False if mdu_physics.temperature == 0 else True,
            "salinity": mdu_physics.salinity,
        }
        return temperature_and_salinity_info


def save_mdu_file(content: List[str], output_path: PathOrStr) -> None:
    """Save the updated MDU file content to disk.

    Args:
        content: The updated content of the MDU file
        output_path: The path where the updated MDU file should be saved
    """
    with open(output_path, "w", encoding="utf-8") as f:
        f.writelines(content)


def get_ref_time(input_date: str, date_format: str = "%Y%m%d"):
    """Convert a date string to a datetime object."""
    date_object = datetime.strptime(f"{input_date}", date_format)
    return f"MINUTES SINCE {date_object}"



import logging
from abc import ABC
from typing import Dict, List, Literal, Optional

from pydantic.v1 import Field
from pydantic.v1.class_validators import root_validator, validator
from pydantic.v1.types import NonNegativeFloat, PositiveInt
from strenum import StrEnum

from hydrolib.core.basemodel import DiskOnlyFileModel
from hydrolib.core.dflowfm.common import LocationType
from hydrolib.core.dflowfm.common.models import Operand
from hydrolib.core.dflowfm.ini.models import INIBasedModel, INIGeneral, INIModel
from hydrolib.core.dflowfm.ini.util import (
    get_enum_validator,
    make_list_validator,
    validate_required_fields,
)

logger = logging.getLogger(__name__)


class DataFileType(StrEnum):
    """
    Enum class containing the valid values for the dataFileType
    attribute in several subclasses of AbstractIniField.
    """

    arcinfo = "arcinfo"
    geotiff = "GeoTIFF"
    sample = "sample"
    onedfield = "1dField"
    polygon = "polygon"
    uniform = "uniform"
    netcdf = "netcdf"

    allowedvaluestext = "Possible values: arcinfo, GeoTIFF, sample, 1dField, polygon."


class InterpolationMethod(StrEnum):
    """
    Enum class containing the valid values for the interpolationMethod
    attribute in several subclasses of AbstractIniField.
    """

    constant = "constant"  # only with dataFileType=polygon .
    triangulation = "triangulation"  # Delaunay triangulation+linear interpolation.
    averaging = "averaging"  # grid cell averaging.
    linear_space_time = "linearSpaceTime"  # linear interpolation in space and time.

    allowedvaluestext = "Possible values: constant, triangulation, averaging."


class AveragingType(StrEnum):
    """
    Enum class containing the valid values for the averagingType
    attribute in several subclasses of AbstractIniField.
    """

    mean = "mean"  # simple average
    nearestnb = "nearestNb"  # nearest neighbour value
    max = "max"  # highest
    min = "min"  # lowest
    invdist = "invDist"  # inverse-weighted distance average
    minabs = "minAbs"  # smallest absolute value

    allowedvaluestext = "Possible values: mean, nearestNb, max, min, invDist, minAbs."


class IniFieldGeneral(INIGeneral):
    """The initial field file's `[General]` section with file meta data."""

    class Comments(INIBasedModel.Comments):
        fileversion: Optional[str] = Field(
            "File version. Do not edit this.", alias="fileVersion"
        )
        filetype: Optional[str] = Field(
            "File type. Should be 'iniField'. Do not edit this.",
            alias="fileType",
        )

    comments: Comments = Comments()
    _header: Literal["General"] = "General"
    fileversion: str = Field("2.00", alias="fileVersion")
    filetype: Literal["iniField"] = Field("iniField", alias="fileType")


class AbstractSpatialField(INIBasedModel, ABC):
    """
    Abstract base class for `[Initial]` and `[Parameter]` block data in
    inifield files.

    Defines all common fields. Used via subclasses InitialField and ParameterField.
    """

    class Comments(INIBasedModel.Comments):
        quantity: Optional[str] = Field(
            "Name of the quantity. See UM Table D.2.", alias="quantity"
        )
        datafile: Optional[str] = Field(
            "Name of file containing field data values.", alias="dataFile"
        )
        datafiletype: Optional[str] = Field("Type of dataFile.", alias="dataFileType")
        interpolationmethod: Optional[str] = Field(
            "Type of (spatial) interpolation.", alias="interpolationmethod"
        )
        operand: Optional[str] = Field(
            "How this data is combined with previous data for the same quantity (if any).",
            alias="operand",
        )
        averagingtype: Optional[str] = Field(
            "Type of averaging, if interpolationMethod=averaging .",
            alias="averagingtype",
        )
        averagingrelsize: Optional[str] = Field(
            "Relative search cell size for averaging.", alias="averagingrelsize"
        )
        averagingnummin: Optional[str] = Field(
            "Minimum number of points in averaging. Must be ≥ 1.",
            alias="averagingnummin",
        )
        averagingpercentile: Optional[str] = Field(
            "Percentile value for which data values to include in averaging. 0.0 means off.",
            alias="averagingpercentile",
        )
        extrapolationmethod: Optional[str] = Field(
            "Option for (spatial) extrapolation.", alias="extrapolationmethod"
        )
        locationtype: Optional[str] = Field(
            "Target location of interpolation.", alias="locationtype"
        )
        value: Optional[str] = Field(
            "Only for dataFileType=polygon. The constant value to be set inside for all model points inside the polygon."
        )

    comments: Comments = Comments()

    quantity: str = Field(alias="quantity")
    datafile: DiskOnlyFileModel = Field(alias="dataFile")

    datafiletype: DataFileType = Field(alias="dataFileType")
    interpolationmethod: Optional[InterpolationMethod] = Field(
        alias="interpolationMethod"
    )
    operand: Optional[Operand] = Field(Operand.override.value, alias="operand")
    averagingtype: Optional[AveragingType] = Field(
        AveragingType.mean.value, alias="averagingType"
    )
    averagingrelsize: Optional[NonNegativeFloat] = Field(1.01, alias="averagingRelSize")
    averagingnummin: Optional[PositiveInt] = Field(1, alias="averagingNumMin")
    averagingpercentile: Optional[NonNegativeFloat] = Field(
        0, alias="averagingPercentile"
    )
    extrapolationmethod: Optional[bool] = Field(False, alias="extrapolationMethod")
    locationtype: Optional[LocationType] = Field(
        LocationType.all.value, alias="locationType"
    )
    value: Optional[float] = Field(alias="value")

    datafiletype_validator = get_enum_validator("datafiletype", enum=DataFileType)
    interpolationmethod_validator = get_enum_validator(
        "interpolationmethod", enum=InterpolationMethod
    )
    operand_validator = get_enum_validator("operand", enum=Operand)
    averagingtype_validator = get_enum_validator("averagingtype", enum=AveragingType)
    locationtype_validator = get_enum_validator("locationtype", enum=LocationType)

    @root_validator(allow_reuse=True)
    def validate_that_value_is_present_for_polygons(cls, values: Dict) -> Dict:
        """Validates that the value is provided when dealing with polygons."""
        return validate_required_fields(
            values,
            "value",
            conditional_field_name="datafiletype",
            conditional_value=DataFileType.polygon,
        )

    @validator("value", always=True)
    @classmethod
    def _validate_value_and_filetype(cls, v, values: dict):
        if v is not None and values.get("datafiletype") != DataFileType.polygon:
            raise ValueError(
                f"When value={v} is given, dataFileType={DataFileType.polygon} is required."
            )

        return v


class InitialField(AbstractSpatialField):
    """
    Initial condition field definition, represents an `[Initial]` block in
    an inifield file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile.initial[..]`

    All lowercased attributes match with the initial field input as described in
    [UM Sec.D.2](https://content.oss.deltares.nl/delft3dfm1d2d/D-Flow_FM_User_Manual_1D2D.pdf#subsection.D.2).
    """

    _header: Literal["Initial"] = "Initial"


class ParameterField(AbstractSpatialField):
    """
    Parameter field definition, represents a `[Parameter]` block in
    an inifield file.
    Typically inside the definition list of a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile.parameter[..]`
    """

    _header: Literal["Parameter"] = "Parameter"


class IniFieldModel(INIModel):
    """
    The overall inifield model that contains the contents of one initial field and parameter file.

    This model is typically referenced under a [FMModel][hydrolib.core.dflowfm.mdu.models.FMModel]`.geometry.inifieldfile[..]`.

    Attributes:
        general (IniFieldGeneral): `[General]` block with file metadata.
        initial (List[InitialField]): List of `[Initial]` blocks with initial condition definitions.
        parameter (List[ParameterField]): List of `[Parameter]` blocks with spatial parameter definitions.
    """

    general: IniFieldGeneral = IniFieldGeneral()
    initial: List[InitialField] = Field(default_factory=list)
    parameter: List[ParameterField] = Field(default_factory=list)

    _split_to_list = make_list_validator("initial", "parameter")

    @classmethod
    def _ext(cls) -> str:
        return ".ini"

    @classmethod
    def _filename(cls) -> str:
        return "fieldFile"



import os
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

from tqdm import tqdm

from hydrolib.core.basemodel import PathOrStr
from hydrolib.core.dflowfm.ext.models import (
    Boundary,
    ExtModel,
    Lateral,
    Meteo,
    SourceSink,
)
from hydrolib.core.dflowfm.extold.models import ExtOldModel
from hydrolib.core.dflowfm.inifield.models import (
    IniFieldModel,
    InitialField,
    ParameterField,
)
from hydrolib.core.dflowfm.structure.models import Structure, StructureModel
from hydrolib.tools.extforce_convert.converters import ConverterFactory
from hydrolib.tools.extforce_convert.mdu_parser import MDUParser
from hydrolib.tools.extforce_convert.utils import (
    backup_file,
    construct_filemodel_new_or_existing,
)


class ExternalForcingConverter:

    def __init__(
        self,
        extold_model: Union[PathOrStr, ExtOldModel],
        ext_file: Optional[PathOrStr] = None,
        inifield_file: Optional[PathOrStr] = None,
        structure_file: Optional[PathOrStr] = None,
        mdu_parser: MDUParser = None,
        verbose: bool = False,
    ):
        """Initialize the converter.

        The converter will create new external forcing, initial field and structure files in the same directory as the
        old external forcing file, if no paths were given by the user for the new models.

        Args:
            extold_model (PathOrStr, ExtOldModel):
                ExtOldModel or path to the old external forcing file.
            ext_file (PathOrStr, optional):
                Path to the new external forcing file.
            inifield_file (PathOrStr, optional):
                Path to the initial field file.
            structure_file (PathOrStr, optional):
                Path to the structure file.
            mdu_parser (Optional[MDUParser], optional):
                a Parser for the MDU file.
            verbose (bool, optional, Defaults to False):
                Enable verbose output.

        Raises:
            FileNotFoundError: If the old external forcing file does not exist.

        Examples:
            >>> converter = ExternalForcingConverter("old-external-forcing.ext") #doctest: +SKIP
            >>> converter.update() #doctest: +SKIP
        """
        if isinstance(extold_model, Path) or isinstance(extold_model, str):
            extold_model = self._read_old_file(extold_model)

        self._extold_model = extold_model
        self._verbose = verbose
        rdir = extold_model.filepath.parent
        self._root_dir = rdir

        # create the new models if not provided by the user in the same directory as the old external file
        path = rdir / "new-external-forcing.ext" if ext_file is None else ext_file
        self._ext_model = construct_filemodel_new_or_existing(ExtModel, path)

        path = (
            rdir / "new-initial-conditions.ini"
            if inifield_file is None
            else inifield_file
        )
        self._inifield_model = construct_filemodel_new_or_existing(IniFieldModel, path)

        path = rdir / "new-structure.ini" if structure_file is None else structure_file
        self._structure_model = construct_filemodel_new_or_existing(
            StructureModel, path
        )

        if mdu_parser is not None:
            self.temperature_salinity_data: Dict[str, int] = (
                mdu_parser.temperature_salinity_data
            )
            self._mdu_parser = mdu_parser

    @property
    def verbose(self) -> bool:
        """bool: Enable verbose output."""
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool):
        self._verbose = value

    @property
    def mdu_parser(self) -> MDUParser:
        if hasattr(self, "_mdu_parser"):
            val = self._mdu_parser
        else:
            val = None
        return val

    @property
    def root_dir(self) -> Path:
        """Root directory of the external forcing file."""
        return self._root_dir

    @property
    def extold_model(self) -> ExtOldModel:
        """old external forcing model."""
        return self._extold_model

    @property
    def ext_model(self) -> ExtModel:
        """New External forcing Model."""
        if not hasattr(self, "_ext_model"):
            raise ValueError(
                "ext_model not set, please use the `ext_model` setter. to set it."
            )
        return self._ext_model

    @ext_model.setter
    def ext_model(self, path: PathOrStr):
        """New external forcing model.

        Args:
            path (PathOrStr, optional): Path to the new external forcing file.
        """
        self._ext_model = construct_filemodel_new_or_existing(ExtModel, path)

    @property
    def inifield_model(self) -> IniFieldModel:
        """IniFieldModel: object with all initial fields blocks."""
        if not hasattr(self, "_inifield_model"):
            raise ValueError(
                "inifield_model not set, please use the `inifield_model` setter. to set it."
            )
        return self._inifield_model

    @inifield_model.setter
    def inifield_model(self, path: PathOrStr):
        self._inifield_model = construct_filemodel_new_or_existing(IniFieldModel, path)

    @property
    def structure_model(self) -> StructureModel:
        """StructureModel: object with all structure blocks."""
        if not hasattr(self, "_structure_model"):
            raise ValueError(
                "structure_model not set, please use the `structure_model` setter. to set it."
            )
        return self._structure_model

    @structure_model.setter
    def structure_model(self, path: PathOrStr):
        self._structure_model = construct_filemodel_new_or_existing(
            StructureModel, path
        )

    @staticmethod
    def _read_old_file(extoldfile: PathOrStr) -> ExtOldModel:
        """Read a legacy D-Flow FM external forcings file (.ext) into an
               ExtOldModel object.

        Args:
            extoldfile (PathOrStr): path to the external forcings file (.ext)

            Returns:
                ExtOldModel: object with all forcing blocks.

            Raises:
                FileNotFoundError: If the old external forcing file does not exist.
        """
        if not isinstance(extoldfile, Path):
            extoldfile = Path(extoldfile)

        if not extoldfile.exists():
            raise FileNotFoundError(f"File not found: {extoldfile}")

        extold_model = ExtOldModel(extoldfile)

        return extold_model

    def update(
        self,
    ) -> Union[Tuple[ExtModel, IniFieldModel, StructureModel], None]:
        """
        Convert the old external forcing file to a new format files.
        When the output files are existing, output will be appended to them.

        Returns:
            Tuple[ExtOldModel, ExtModel, IniFieldModel, StructureModel]:
                The updated models (already written to disk). Maybe used
                at call site to inspect the updated models.
        """
        self._log_conversion_details()

        for forcing in self.extold_model.forcing:

            new_quantity_block = self._convert_forcing(forcing)

            if isinstance(new_quantity_block, Boundary):
                self.ext_model.boundary.append(new_quantity_block)
            elif isinstance(new_quantity_block, Lateral):
                self.ext_model.lateral.append(new_quantity_block)
            elif isinstance(new_quantity_block, SourceSink):
                self.ext_model.sourcesink.append(new_quantity_block)
            elif isinstance(new_quantity_block, Meteo):
                self.ext_model.meteo.append(new_quantity_block)
            elif isinstance(new_quantity_block, InitialField):
                self.inifield_model.initial.append(new_quantity_block)
            elif isinstance(new_quantity_block, ParameterField):
                self.inifield_model.parameter.append(new_quantity_block)
            elif isinstance(new_quantity_block, Structure):
                self.structure_model.structure.append(new_quantity_block)
            else:
                raise NotImplementedError(
                    f"Unsupported model class {type(new_quantity_block)} for {forcing.quantity} in "
                    f"{self.extold_model.filepath}."
                )

        if self.mdu_parser is not None:
            self._update_mdu_file()

        return self.ext_model, self.inifield_model, self.structure_model

    def _convert_forcing(self, forcing) -> Union[Boundary, Lateral, Meteo, SourceSink]:
        """Convert a single forcing block to the appropriate new format.

        Notes:
            - The SourceSink converter needs the salinity and temperature from the FM model.
            - The BoundaryCondition converter needs the start time from the FM model.
        """

        converter_class = ConverterFactory.create_converter(forcing.quantity)
        converter_class.root_dir = self.root_dir

        # only the SourceSink converter needs the quantities' list
        if converter_class.__class__.__name__ == "SourceSinkConverter":

            if self.temperature_salinity_data is None:
                raise ValueError(
                    "FM model is required to convert SourcesSink quantities."
                )
            else:
                temp_salinity_mdu = self.temperature_salinity_data
                start_time = self.temperature_salinity_data.get("refdate")

            quantities = self.extold_model.quantities
            new_quantity_block = converter_class.convert(
                forcing, quantities, start_time=start_time, **temp_salinity_mdu
            )
        elif converter_class.__class__.__name__ == "BoundaryConditionConverter":
            if self.temperature_salinity_data is None:
                raise ValueError("FM model is required to convert Boundary conditions.")
            else:
                start_time = self.temperature_salinity_data.get("refdate")
                new_quantity_block = converter_class.convert(forcing, start_time)
        else:
            new_quantity_block = converter_class.convert(forcing)

        return new_quantity_block

    def save(self, backup: bool = True, recursive: bool = True):
        """Save the updated models to disk.

        Args:
            backup (bool, optional):
                Create a backup of each file that will be overwritten.
            recursive (bool, optional): Defaults to True.
                Save the models recursively.
        """
        if (
            len(self.inifield_model.parameter) > 0
            or len(self.inifield_model.initial) > 0
        ):
            self._save_inifield_model(backup, recursive)

        if len(self.structure_model.structure) > 0:
            self._save_structure_model(backup, recursive)

        num_quantities_ext = (
            len(self.ext_model.meteo)
            + len(self.ext_model.sourcesink)
            + len(self.ext_model.boundary)
            + len(self.ext_model.lateral)
        )
        if num_quantities_ext:
            if backup and self.ext_model.filepath.exists():
                backup_file(self.ext_model.filepath)
            self.ext_model.save(recurse=recursive)

        if self.mdu_parser is not None:
            self.mdu_parser.save(backup=True)

    def _save_inifield_model(self, backup: bool, recursive: bool):
        if backup and self.inifield_model.filepath.exists():
            backup_file(self.inifield_model.filepath)
        self.inifield_model.save(recurse=recursive)

    def _save_structure_model(self, backup: bool, recursive: bool):
        if backup and self.structure_model.filepath.exists():
            backup_file(self.structure_model.filepath)
        self.structure_model.save(recurse=recursive)

    def clean(self):
        """
        Clean the directory from the old external forcing file and the time file.
        """
        root_dir = self.root_dir
        time_files = list(root_dir.glob("*.tim"))
        if len(time_files) > 0:
            for file in time_files:
                file.unlink()

        self.extold_model.filepath.unlink()

    @classmethod
    def from_mdu(
        cls,
        mdu_file: PathOrStr,
        ext_file: Optional[PathOrStr] = None,
        inifield_file: Optional[PathOrStr] = None,
        structure_file: Optional[PathOrStr] = None,
    ) -> "ExternalForcingConverter":
        """class method to create the converter from MDU file.

        Args:
            mdu_file (PathOrStr): Path to the D-Flow FM main input file (.mdu).
                Must be parsable into a standard FMModel.
                When this contains a valid filename for ExtFile, conversion
                will be performed.
            ext_file (PathOrStr, optional): Path to the output external forcings
                file. Defaults to the given ExtForceFileNew in the MDU file, if
                present, or forcings.ext otherwise.
            inifield_file (PathOrStr, optional): Path to the output initial field
                file. Defaults to the given IniFieldFile in the MDU file, if
                present, or inifields.ini otherwise.
            structure_file (PathOrStr, optional): Path to the output structures.ini
                file. Defaults to the given StructureFile in the MDU file, if
                present, or structures.ini otherwise.

        Returns:
            ExternalForcingConverter: The converter object.

        Raises:
            FileNotFoundError: If the MDU file does not exist.
            ValueError: If the old external forcing file is not found in the MDU file.
            DeprecationWarning: If the MDU file contains unknown keywords.
        """
        if isinstance(mdu_file, str):
            mdu_file = Path(mdu_file)

        mdu_parser = MDUParser(mdu_file)

        external_forcing_data = mdu_parser.loaded_fm_data.get("external_forcing")
        geometry = mdu_parser.loaded_fm_data.get("geometry")

        inifieldfile_mdu = geometry.get("inifieldfile")
        structurefile_mdu = geometry.get("structurefile")

        old_ext_force_file = external_forcing_data.get("extforcefile")
        if old_ext_force_file is None:
            raise ValueError(
                "An old formatted external forcing file (.ext) "
                "could not be found in the mdu file.\n"
                "Conversion is not possible or may not be necessary."
            )

        new_ext_force_file = external_forcing_data.get("extforcefilenew")

        old_ext_force_file = (
            Path(old_ext_force_file)
            if old_ext_force_file is not None
            else old_ext_force_file
        )
        new_ext_force_file = (
            Path(new_ext_force_file)
            if new_ext_force_file is not None
            else new_ext_force_file
        )

        root_dir = mdu_file.parent
        extoldfile = root_dir / old_ext_force_file

        if new_ext_force_file:
            ext_file = new_ext_force_file.resolve()
        else:
            if ext_file is None:
                old_ext = old_ext_force_file.with_stem(old_ext_force_file.stem + "-new")
                ext_file = root_dir / old_ext
            else:
                ext_file = root_dir / ext_file

        inifield_file = ExternalForcingConverter._get_inifield_file(
            inifield_file, root_dir, inifieldfile_mdu
        )
        structure_file = ExternalForcingConverter._get_structure_file(
            structure_file, root_dir, structurefile_mdu
        )

        return cls(extoldfile, ext_file, inifield_file, structure_file, mdu_parser)

    @staticmethod
    def _get_inifield_file(
        inifield_file: Optional[PathOrStr],
        root_dir: Path,
        inifieldfile_mdu: Optional[PathOrStr],
    ) -> Path:
        if inifield_file is not None:
            # user defined initial field file
            inifield_file = root_dir / inifield_file
        elif isinstance(inifieldfile_mdu, Path):
            # from the LegacyFMModel
            inifield_file = inifieldfile_mdu.resolve()
        elif isinstance(inifieldfile_mdu, str):
            # from reading the geometry section
            inifield_file = root_dir / inifieldfile_mdu
        else:
            print(
                f"The initial field file is not found in the mdu file, and not provided by the user. \n "
                f"given: {inifield_file}."
            )
        return inifield_file

    @staticmethod
    def _get_structure_file(
        structure_file: Optional[PathOrStr],
        root_dir: Path,
        structurefile_mdu: Optional[PathOrStr],
    ) -> Path:
        if structure_file is not None:
            # user defined structure file
            structure_file = root_dir / structure_file
        elif isinstance(structurefile_mdu, Path):
            # from the LegacyFMModel
            structure_file = structurefile_mdu.resolve()
        elif isinstance(structurefile_mdu, str):
            # from reading the geometry section
            structure_file = root_dir / structurefile_mdu
        else:
            print(
                "The structure file is not found in the mdu file, and not provide by the user. \n"
                f"given: {structure_file}."
            )
        return structure_file

    def _update_mdu_file(self):
        """Update the FM model with the new external forcings, initial fields and structures files.

        - The FM model will be saved with a postfix added to the filename.
        - The original FM model will be backed up.

        Notes:
            -If the `fm_model` was not read correctly due to `Unknown keywords` the function will update the field of the
            `ExtForceFileNew` in the mdu file, and store the new content in the `mdu_info` dictionary under a
            `new_mdu_content` key.
        """
        new_ext_file = self.ext_model.filepath.name
        self.mdu_parser.new_forcing_file = new_ext_file
        self.mdu_parser.content = self.mdu_parser.update_extforce_file_new()

    def _log_conversion_details(self):
        """Log details about the conversion process if verbosity is enabled."""
        if self.verbose:
            workdir = os.getcwd() + "\\"
            print(f"Work dir: {workdir}")
            print("Using attribute files:")
            print("Input:")
            print(f"* {self.extold_model.filepath}")
            print("Output:")
            print(f"* {self.ext_model.filepath}")
            print(f"* {self.inifield_model.filepath}")
            print(f"* {self.structure_model.filepath}")


def recursive_converter(
    root_dir: PathOrStr,
    backup: bool = True,
    suppress_errors: bool = False,
    remove_legacy: bool = False,
):
    """Migrate all external forcings files in a directory tree to the new format.

    Args:
        root_dir: Directory to recursively find and convert .mdu files in.
        backup (bool, optional): Create a backup of each file that will be overwritten.
        suppress_errors (bool, optional): Suppress errors during conversion.
        remove_legacy (bool, optional): Remove legacy/old files (e.g. .tim) after conversion. Defaults to False.
    """
    mdu_files = [
        path for path in Path(root_dir).rglob("*.mdu") if "_ext" not in path.name
    ]

    if not mdu_files:
        print("No .mdu files found in the specified directory.")
    else:
        print(f"Found {len(mdu_files)} .mdu files. Starting conversion...")

        for path in tqdm(mdu_files, desc="Converting files"):
            try:
                converter = ExternalForcingConverter.from_mdu(path)
                _, _, _ = converter.update()
                converter.save(backup=backup)
                if remove_legacy:
                    converter.clean()
            except Exception as e:
                if not suppress_errors:
                    print(f"Error processing {path}: {e}")



"""CLI.

CLI for converting D-Flow FM legacy external forcings files to current external forcings file/initial
fields file/structures file.
"""

import argparse
from argparse import ArgumentTypeError, Namespace
from pathlib import Path

from hydrolib.core import __version__
from hydrolib.tools.extforce_convert.main_converter import (
    ExternalForcingConverter,
    recursive_converter,
)


def valid_file(path_str):
    """Check if the file exists and return its path."""
    path = Path(path_str)
    if not path.exists():
        raise ArgumentTypeError(f"File not found: {path}")
    return path


def _get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="extforce_convert",
        description="Convert D-Flow FM legacy external forcings files to current external forcings file/initial fields file/structures file.",
    )

    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Print diagnostic information"
    )

    # mdu file, extforcefile and dir are mutually exclusive (can only use one)
    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--mdufile",
        "-m",
        action="store",
        type=valid_file,
        help="Automatically take input and output filenames from MDUFILE",
    )
    group.add_argument(
        "--extoldfile",
        "-e",
        action="store",
        type=valid_file,
        help="Input EXTOLDFILE to be converted.",
    )
    group.add_argument(
        "--dir",
        "-d",
        action="store",
        help="Directory to recursively find and convert .mdu files in",
    )

    parser.add_argument(
        "--outfiles",
        "-o",
        action="store",
        nargs=3,
        metavar=("EXTFILE", "INIFIELDFILE", "STRUCTUREFILE"),
        help="Save forcings, initial fields and structures to specified filenames",
    )

    backup_group = parser.add_mutually_exclusive_group()

    backup_group.add_argument(
        "--backup",
        "-b",
        action="store_true",
        default=True,
        help="Create a backup of each file that will be overwritten.",
    )
    backup_group.add_argument(
        "--no-backup",
        dest="backup",
        action="store_false",
        help="Do not create a backup of each file that will be overwritten.",
    )

    parser.add_argument(
        "--remove-legacy-files",
        dest="remove_legacy",
        action="store_true",
        default=False,
        help="Remove legacy/old files (e.g. .tim) after conversion. Defaults to False.",
    )
    return parser


def main(args=None):
    """Entry point for extforce_convert tool.

    Args:
        args : list
            A of arguments as if they were input in the command line. Leave it
            None to use sys.argv.
    """
    parser = _get_parser()
    args = parser.parse_args(args)

    # three cases to consider
    if args.mdufile:
        convert_with_mdu_file(args)
    elif args.extoldfile is not None:
        convert_with_extold_file(args)
    elif args.dir is not None:
        recursive_converter(
            args.dir, backup=args.backup, remove_legacy=args.remove_legacy
        )
    else:
        print("Error: no input specified. Use one of --mdufile, --extoldfile or --dir.")


def convert_with_mdu_file(args: Namespace):
    """Convert the old external forcing file using the mdu file.

    Read the old external forcing file path from the mdu file,
    and convert it to the new format files.

    Args:
        args : Namespace
            The arguments parsed from the command line.
    """
    converter = ExternalForcingConverter.from_mdu(
        args.mdufile,
        ext_file=(args.outfiles[0] if args.outfiles else None),
        inifield_file=(args.outfiles[1] if args.outfiles else None),
        structure_file=(args.outfiles[2] if args.outfiles else None),
    )
    convert(converter, args)


def convert_with_extold_file(args: Namespace):
    """Convert the old external forcing file to the new format files.

    Args:
        args : Namespace
            The arguments parsed from the command line.
    """
    converter = ExternalForcingConverter(
        args.extoldfile,
        ext_file=(args.outfiles[0] if args.outfiles else None),
        inifield_file=(args.outfiles[1] if args.outfiles else None),
        structure_file=(args.outfiles[2] if args.outfiles else None),
    )
    convert(converter, args)


def convert(converter: ExternalForcingConverter, args: Namespace):
    """Run the update method of ExternalForcingConverter and save the results.

    Args:
        converter : ExternalForcingConverter
            The converter object to convert the old external forcing file.
        args : Namespace
            The arguments parsed from the command line.
    """
    converter.verbose = args.verbose
    converter.update()
    print("Converting the old external forcing file to the new format files is done.")
    converter.save(backup=args.backup)
    print("The new files are saved.")
    if args.remove_legacy:
        print("Cleaning legacy tim files ...")
        converter.clean()


if __name__ == "__main__":
    main()



from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, Type, Union

from pydantic.v1 import Extra

from hydrolib.core.basemodel import DiskOnlyFileModel, FileModel, PathOrStr
from hydrolib.core.dflowfm.ext.models import (
    MeteoForcingFileType,
    MeteoInterpolationMethod,
)
from hydrolib.core.dflowfm.extold.models import ExtOldFileType, ExtOldForcing
from hydrolib.core.dflowfm.inifield.models import (
    AveragingType,
    DataFileType,
    InterpolationMethod,
)


def construct_filemodel_new_or_existing(
    model_class: Type[FileModel], filepath: PathOrStr, *args, **kwargs
) -> FileModel:
    """Construct a new instance of a FileModel subclass, either loading an
    existing model file, or starting a blank one.

    If the given filepath exists, it will be loaded, as if the constructor
    was directly called. If it does not exist, the blank model instance
    will only have its filepath attribute set, e.g., for future saving.

    Args:
        model_class (type[FileModel]): The FileModel subclass to construct or update.
        filepath (PathOrStr): The filepath to use for the new or existing model input.
        *args: Additional positional arguments to pass to the ModelClass constructor
        **kwargs: Additional keywords arguments to pass to the ModelClass constructor
    """
    if Path(filepath).is_file():
        model = model_class(filepath=filepath, *args, **kwargs)
    else:
        model = model_class(*args, **kwargs)
        model.filepath = filepath

    return model


def backup_file(filepath: PathOrStr) -> None:
    """Create a backup of the given file by copying it to a new file with a
    '.bak' extension.

    Args:
        filepath (PathOrStr): The path to the file to back up.
    """
    filepath = Path(filepath) if isinstance(filepath, str) else filepath
    if filepath.is_file():
        backup_path = filepath.with_suffix(filepath.suffix + ".bak")
        filepath.replace(backup_path)


def construct_filepath_with_postfix(filepath: PathOrStr, postfix: str) -> Path:
    """Construct a new filepath by adding a postfix to the existing
    filepath, still keeping the original file suffix.
    For example, 'file.txt' with postfix '_new' will become 'file_new.txt'.

    Args:
        filepath (PathOrStr): The path to the file.
        postfix (str): The postfix to add to the filename.

    Returns:
        Path: The new filepath with the postfix included.

    Examples:
        ```python
        >>> construct_filepath_with_postfix("file.txt", "_new") # doctest: +SKIP
        Path("file_new.txt")
        ```
    """
    file_as_path = Path(filepath)
    return file_as_path.with_stem(file_as_path.stem + postfix)


def oldfiletype_to_forcing_file_type(
    oldfiletype: int,
) -> Union[MeteoForcingFileType, str]:
    """Convert old external forcing `FILETYPE` integer value to valid
        `forcingFileType` string value.

    Args:
        oldfiletype (int): The FILETYPE value in an old external forcings file.

    Returns:
        Union[MeteoForcingFileType,str]: Corresponding value for `forcingFileType`,
            or "unknown" for invalid input.
    """

    forcing_file_type = "unknown"

    if oldfiletype == ExtOldFileType.TimeSeries:  # 1
        forcing_file_type = MeteoForcingFileType.uniform
    elif oldfiletype == ExtOldFileType.TimeSeriesMagnitudeAndDirection:  # 2
        forcing_file_type = MeteoForcingFileType.unimagdir
    elif oldfiletype == ExtOldFileType.SpatiallyVaryingWindPressure:  # 3
        raise NotImplementedError(
            "FILETYPE = 3 (spatially verying wind and pressure) is no longer supported."
        )
    elif oldfiletype == ExtOldFileType.ArcInfo:  # 4
        forcing_file_type = MeteoForcingFileType.arcinfo
    elif oldfiletype == ExtOldFileType.SpiderWebData:  # 5
        forcing_file_type = MeteoForcingFileType.spiderweb
    elif oldfiletype == ExtOldFileType.CurvilinearData:  # 6
        forcing_file_type = MeteoForcingFileType.curvigrid
    elif oldfiletype == ExtOldFileType.Samples:  # 7
        forcing_file_type = DataFileType.sample
    elif oldfiletype == ExtOldFileType.TriangulationMagnitudeAndDirection:  # 8
        raise NotImplementedError(
            "FILETYPE = 8 (magnitude+direction timeseries on stations) is no longer supported."
        )
    elif oldfiletype == ExtOldFileType.Polyline:  # 9
        # Boundary polyline files no longer need a filetype of their own (intentionally no error raised)
        pass
    elif oldfiletype == ExtOldFileType.InsidePolygon:  # 10
        forcing_file_type = DataFileType.polygon
    elif oldfiletype == ExtOldFileType.NetCDFGridData:  # 11
        forcing_file_type = MeteoForcingFileType.netcdf

    return forcing_file_type


def oldmethod_to_interpolation_method(
    oldmethod: int,
) -> Union[InterpolationMethod, MeteoInterpolationMethod, str]:
    """Convert old external forcing `METHOD` integer value to valid
        `interpolationMethod` string value.

    Args:
        oldmethod (int): The METHOD value in an old external forcings file.

    Returns:
        Union[InterpolationMethod,str]: Corresponding value for `interpolationMethod`,
            or "unknown" for invalid input.
    """

    if oldmethod in [1, 2, 3, 11]:
        interpolation_method = MeteoInterpolationMethod.linearSpaceTime
    elif oldmethod == 5:
        interpolation_method = InterpolationMethod.triangulation
    elif oldmethod == 4:
        interpolation_method = InterpolationMethod.constant
    elif oldmethod in range(6, 10):
        interpolation_method = InterpolationMethod.averaging
    else:
        interpolation_method = "unknown"
    return interpolation_method


def oldmethod_to_averaging_type(
    oldmethod: int,
) -> Union[AveragingType, str]:
    """Convert old external forcing `METHOD` integer value to valid
        `averagingType` string value.

    Args:
        oldmethod (int): The METHOD value in an old external forcings file.

    Returns:
        Union[AveragingType,str]: Corresponding value for `averagingType`,
            or "unknown" for invalid input.
    """

    if oldmethod == 6:
        averaging_type = AveragingType.mean
    else:
        averaging_type = "unknown"

    return averaging_type


def convert_interpolation_data(
    forcing: ExtOldForcing, data: Dict[str, Any]
) -> Dict[str, str]:
    """Convert interpolation data from old to new format.

    Args:
        forcing (ExtOldForcing): The old forcing block with interpolation data.
        data (Dict[str, Any]): The dictionary to which the new data will be added.

    Returns:
        Dict[str, str]: The updated dictionary with the new interpolation data.
        - The dictionary will contain the "interpolationmethod" key with the new interpolation method.
        - if the interpolation method is "Averaging" (method = 6), the dictionary will also contain
            the "averagingtype", "averagingrelsize", "averagingnummin", and "averagingpercentile" keys.
    """
    data["interpolationmethod"] = oldmethod_to_interpolation_method(forcing.method)
    if data["interpolationmethod"] == InterpolationMethod.averaging:
        data["averagingtype"] = oldmethod_to_averaging_type(forcing.method)
        data["averagingrelsize"] = forcing.relativesearchcellsize
        data["averagingnummin"] = forcing.nummin
        data["averagingpercentile"] = forcing.percentileminmax

    return data


def create_initial_cond_and_parameter_input_dict(
    forcing: ExtOldForcing,
) -> Dict[str, str]:
    """Create the input dictionary for the `InitialField` or `ParameterField`

    Args:
        forcing: [ExtOldForcing]
            External forcing block from the old external forcings file.

    Returns:
        Dict[str, str]:
            the input dictionary to the `InitialField` or `ParameterField` constructor
    """
    block_data = {
        "quantity": forcing.quantity,
        "datafile": DiskOnlyFileModel(forcing.filename.filepath),
        "datafiletype": oldfiletype_to_forcing_file_type(forcing.filetype),
    }
    if block_data["datafiletype"] == "polygon":
        block_data["value"] = forcing.value

    if forcing.sourcemask != DiskOnlyFileModel(None):
        raise ValueError(
            f"Attribute 'SOURCEMASK' is no longer supported, cannot "
            f"convert this input. Encountered for QUANTITY="
            f"{forcing.quantity} and FILENAME={forcing.filename}."
        )
    block_data = convert_interpolation_data(forcing, block_data)
    block_data["operand"] = forcing.operand

    if hasattr(forcing, "extrapolation"):
        block_data["extrapolationmethod"] = (
            "yes" if forcing.extrapolation == 1 else "no"
        )

    return block_data


def find_temperature_salinity_in_quantities(strings: List[str]) -> Dict[str, int]:
    """
    Searches for keywords "temperature" and "salinity" in a list of strings
    and returns a dictionary with associated values.

    Args:
        strings (List[str]): A list of strings to search.

    Returns:
        Dict[str, int]: A dictionary with keys as "salinity" or "temperature"
                        and values 3 and 4 respectively.

     Examples:
         ```python
        >>> find_temperature_salinity_in_quantities(["temperature", "Salinity"])
        OrderedDict({'salinitydelta': 3, 'temperaturedelta': 4})
        >>> find_temperature_salinity_in_quantities(["Temperature"])
        OrderedDict({'temperaturedelta': 3})
        >>> find_temperature_salinity_in_quantities(["Salinity"])
        OrderedDict({'salinitydelta': 3})
        >>> find_temperature_salinity_in_quantities(["tracers"])
        OrderedDict()
        >>> find_temperature_salinity_in_quantities([])
        OrderedDict()
        >>> find_temperature_salinity_in_quantities(["discharge_salinity_temperature_sorsin", "Salinity"])
        OrderedDict({'salinitydelta': 3})

        ```

    Notes:
        - The function removes the `discharge_salinity_temperature_sorsin` from the given list of strings.
        - The function removes the duplicate strings in the list.
    """
    result = OrderedDict()
    strings = list(set(strings))
    # remove the `discharge_salinity_temperature_sorsin` quantity from the list
    if "discharge_salinity_temperature_sorsin" in strings:
        strings.remove("discharge_salinity_temperature_sorsin")

    if any("salinity" in string.lower() for string in strings):
        result["salinitydelta"] = 3
    if any("temperature" in string.lower() for string in strings):
        result["temperaturedelta"] = (
            result.get("salinitydelta", 2) + 1
        )  # Default temperature value is 2

    return result


class IgnoreUnknownKeyWord(type):
    """Metaclass to ignore unknown keyword arguments when creating a new class instance"""

    def __call__(cls, base_class, **data):
        """Dynamically create and instantiate a subclass of base_class."""

        class DynamicClass(base_class):
            class Config:
                extra = Extra.ignore

            def __init__(self, **data):
                valid_fields = self.__annotations__.keys()
                filtered_data = {k: v for k, v in data.items() if k in valid_fields}
                super().__init__(**filtered_data)

        return DynamicClass(**data)


class IgnoreUnknownKeyWordClass(metaclass=IgnoreUnknownKeyWord):
    pass
